/*
 *  userial -- a USB to I2C converter
 *
 *  Copyright (c) 2008-2009, Thomas Pircher <tehpeh@gmx.net>
 *  
 *  Permission is hereby granted, free of charge, to any person obtaining a copy
 *  of this software and associated documentation files (the "Software"), to deal
 *  in the Software without restriction, including without limitation the rights
 *  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 *  copies of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *  
 *  The above copyright notice and this permission notice shall be included in
 *  all copies or substantial portions of the Software.
 *  
 *  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 *  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 *  THE SOFTWARE.
 *****************************************************************************/
#include "hardware.h"
#include <stdint.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/wdt.h>

#define HW_LED      PORTD   // the LED register
#define HW_LED_DDR  DDRD    // the LED direction register


static inline void led_init(void);


void hw_init(void)
{
    // Disable watchdog if enabled by bootloader/fuses
    MCUSR &= ~(1U << WDRF);
    wdt_disable();
    hw_set_sysclock_prescaler(0);
    SMCR = _BV(SM1) | _BV(SM0);     // Power save mode

    led_init();
}

void hw_enable_int(void)
{
    sei();
}

void hw_delay_ms(uint16_t ms)
{
    const uint16_t loops_per_ms = F_CPU / 4000U; // 4 ticks per inner loop
    uint16_t tmp;
    __asm__ __volatile__(
        "\n1:"                  "\n\t"
        "movw %[tmp], %[lpm]"   "\n\t"
        "\n2:"                  "\n\t"
        "sbiw %[tmp], 1"        "\n\t"
        "brne 2b"               "\n\t"
        "sbiw %[ms], 1"         "\n\t"
        "brne 1b"               "\n\t"
        "clr r1"                "\n\t"
        : [tmp] "=e" (tmp),
          [ms] "+e" (ms)
        : [lpm] "r" (loops_per_ms)
        : "r0"
    );
}

void hw_set_sysclock_prescaler(uint8_t mask)
{
    uint8_t tmp = _BV(CLKPCE);
    __asm__ __volatile__ (
            "in __tmp_reg__,__SREG__" "\n\t"
            "cli" "\n\t"
            "sts %[cpr], %[pce]" "\n\t"
            "sts %[cpr], %[msk]" "\n\t"
            "out __SREG__, __tmp_reg__"
            : // no outputs
            : [pce] "d" (tmp),
              [cpr] "M" (_SFR_MEM_ADDR(CLKPR)),
              [msk] "d" (mask)
            : "r0");
}

void hw_sleep(void)
{
    __asm__ __volatile__("sleep");
}

void led_init(void)
{
    HW_LED &= ~0xc0;
    HW_LED_DDR |= 0xc0;
}

void led_on(led_t led)
{
    HW_LED |= led;
}

void led_off(led_t led)
{
    HW_LED &= ~led;
}

void led_toggle(led_t led)
{
    HW_LED ^= led;
}

